==> ChangePointDetectorEVT-0.0.16.tar.gz: ChangePointDetectorEVT-0.0.16/src/ChangePointDetector/ChangePointDetector.py <==
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 17:56:11 2021
v0.8 28Oct21

@author: Michael Hauptman

CHANGE POINT DETECTOR

This module takes a time series and returns the uunderlying linear trend and changpoints. It uses EVT as described in https://www.robots.ox.ac.uk/~sjrob/Pubs/LeeRoberts_EVT.pdf

INSTRUCTIONS:

1. from ChangePointDetector import ChangePointDetector 
2. Prepare your time series as data plus Panda dates
3. Create  the necessary Kalman representation by creating a "session" object by calling the ChangePoint class, e.g.:
	Session=ChangePointDetector.ChangePointDetectorSession(data,dates). 
    - 'SeasonalityPeriods' is an optional input, e.g 12 = calendar month seasonality
    - 'ForecastPeriods' is another optional input, indicating how many periods to forecast.  Default = 3
4. Determine the changepoints by running the ChangePointDetectorFunction on your "session", e.g.
	Results=Session.ChangePointDetectorFunction()
5. This will return a "Results" object that contains the following:
	- ChangePoints.  This is a list of 0s and 1s the length of the data, where 1s represent changepoints
	- Prediction.  This is the Kalman smoothed actuals, plus a 3 period forecast.  Note no forecast will be made if there is a changepoint in the last 3 		dates
	- PredictionVariance
	- ExtendedDates.  These are the original dates plus 3 exta for the forecast (if a forecast has been made)
	- Trend.  This is the linear change factor
	- TrendVariance.  Variance of the trend


"""

import numpy as np
from numpy.linalg import inv
from statsmodels.tsa.statespace.mlemodel import MLEModel
# import math
from math import sqrt, log, exp,pi
from scipy.stats.distributions import chi2
from dateutil.relativedelta import relativedelta

class ModuleResults:
    def __init__(self,Trend,TrendVariance,ChangePoints,Prediction,PredictionVariance,ExtendedDates):
            self.Trend=Trend
            self.TrendVariance=TrendVariance
            self.ChangePoints=ChangePoints
            self.Prediction=Prediction
            self.ExtendedDates=ExtendedDates
            self.PredictionVariance=PredictionVariance

class TimeSeriesData:
        def __init__(self,data,dates):
            self.data=data
            self.dates=dates
            #Determine periodicity
            Delta=(dates[1]-dates[0])
            if Delta in range(27,32):
                Period = "months"
            elif Delta ==1:
                Period = 'days'
            elif Delta ==7:
                Period = 'weeks'
            else:
                Period ="months"
            self.Period=Period

            
class SeasonalStateArrays:
    def __init__(self,SeasonalityPeriods):
        if SeasonalityPeriods ==0:
            A=np.diag(np.ones(1))
            H=np.ones(1)
        elif SeasonalityPeriods ==1:
            A=np.array([[1,1],[0,1]])
            H=np.array([1,0])
        else:
            A=np.diag(np.ones(SeasonalityPeriods+1))
            B=np.zeros((SeasonalityPeriods+1,1))
            C=-1*np.ones(SeasonalityPeriods+2)
            C[0]=0
            C[-1]=0
            A=np.hstack([A,B])
            A[0,-1]=1
            D=np.zeros(SeasonalityPeriods+2)
            D[-1]=1
            A[-1,:]=D
            A=np.vstack([A[0,:],C,A[1:,:]]).astype('float')
            H=np.zeros((SeasonalityPeriods+2))
            H[0]=1
            H[1]=1
        
        self.transition=A
        self.design=np.array([H])
        
class KalmanModel(MLEModel):
    def __init__(self, endog, transition,design,**dates):
        super().__init__(endog, k_states = transition.shape[0], **dates)

        self['transition']=transition
        self['design']=design
        self['selection']=np.eye(transition.shape[0])

        self.initialize_approximate_diffuse()

    def update(self, params, **kwargs):
        params = super().update(params, **kwargs)
        self['state_cov']=np.eye(self.k_states)
        for i in range(self.k_states):
            self['state_cov'][i,i]=params[0]
        for j in range(self['design'].shape[0]):
            self['obs_cov',j, j] = params[1] ** 2
            
def kgain(P, H, R):
    IS =  H @ P @ H.T + R # the Covariance of predictive mean of Y
    K = P @ H.T @ inv(IS) #State covariance in observation space divided by  Covariance of predictive mean of Y
    return K

def MDCalc(X,Xold,P):
        return(
        sqrt((X-Xold).T@inv(P)@(X-Xold))  #Mahalanobis distance
        )
    
def MD_ProbCalc(MD,DF):
    return(
    chi2.cdf(MD, df=DF)
    )

class MyKalmanFilter:
    def __init__(self, TS_Data, SeasonalityPeriods=1):
        OriginalData=TS_Data.data
        NormalisedData=OriginalData/np.linalg.norm(OriginalData) 
#         NormalisedData=RemoveOutliers(OriginalData)/np.linalg.norm(RemoveOutliers(OriginalData))
        self.OriginalData=OriginalData
        self.data=NormalisedData
        self.dates=TS_Data.dates
        self.SeasonalityPeriods=SeasonalityPeriods
        self.StateArrays=SeasonalStateArrays(SeasonalityPeriods)
            
        
    def InitialiseFilter(self):
        NormalisedData=self.data
        MeasureVar=.01   #Doesn't seem to do anything??
        kf_model=KalmanModel(NormalisedData,transition=self.StateArrays.transition,design=self.StateArrays.design)
        
        start_params = [np.var(NormalisedData)**3, MeasureVar] #State convariance and observation covariance. These settings will underfit the curve
        kf_model_fit = kf_model.fit(start_params,maxiter = 20,method = 'bfgs', hessian= 'true')
        Growth=kf_model_fit.predict(1,len(self.data))
        Base=np.mean(abs(NormalisedData-Growth))

        Success=False
        i=3
        OldRatio=0
        while Success ==False and i>0:  #Increae state covariance until fitting improves over underfit base case
            start_params = [np.var(NormalisedData)**(i), MeasureVar]
            kf_model_fit = kf_model.fit(start_params,maxiter = 20,method = 'bfgs', hessian= 'true')
            Growth=kf_model_fit.predict(1,len(self.data))
            Ratio=np.mean(abs(NormalisedData-Growth))/Base
            if Ratio<.3 or Ratio==OldRatio:  
                Success =True
            i-=0.05
            OldRatio=Ratio
        
        #Determine kalman gain
#         K_gain=np.zeros(self.SeasonalityPeriods+1)
#         for i in range(kf_model_fit.filtered_state.shape[1]):
#             P=kf_model_fit.filtered_state_cov[:,:,i]
#             KGain=kgain(P, kf_model['design'], kf_model['obs_cov'])
#             K_gain=np.vstack([K_gain,KGain.reshape([1,self.SeasonalityPeriods+1])[0]])
            
        self.KalmanModel=kf_model
        self.KalmanModelFit=kf_model_fit
#         self.K_gain=K_gain
        

   
class ChangePointDetectorSession:
    def __init__(self, data,dates, SeasonalityPeriods=1,ForecastPeriods=3):
        self.ts=TimeSeriesData(data,dates)
        self.TimeseriesModel = MyKalmanFilter(self.ts,SeasonalityPeriods)
        self.TimeseriesModel.InitialiseFilter()
        self.SeasonalityPeriods=SeasonalityPeriods
        self.ForecastPeriods=ForecastPeriods
        
    
    def ChangePointDetectorFunction(self):
    #This is the main function    
        #Get underlying trend
#       gs_orig=self.TimeseriesModel.KalmanModelFit.filtered_state[-1,:]
        gs_orig=self.TimeseriesModel.KalmanModelFit.smoothed_state[-1,:]
        gs_orig= np.insert(gs_orig[:-1], 0, 0, axis=0)
        
        gs=TimeSeriesData(gs_orig,self.ts.dates)
        OnePeriodRegressionModel=MyKalmanFilter(gs,SeasonalityPeriods=0)
        OnePeriodRegressionModel.InitialiseFilter()
       
        
        MDout,MD_Prob = [],[]
        DF=1
        for i in range(OnePeriodRegressionModel.KalmanModelFit.filtered_state.shape[1]):
        
            X=OnePeriodRegressionModel.KalmanModelFit.filtered_state[:,i]
            P=OnePeriodRegressionModel.KalmanModelFit.filtered_state_cov[:,:,i]
            if i==0:
                Xold=X
            else:
                Xold=OnePeriodRegressionModel.KalmanModelFit.filtered_state[:,i-1]
            
            #If variance = 0, adjust to be >0 else MD cannot be calculated
            if P[0][0]==0:
                PR=X[0]/1000
            else:
                PR=P[0][0]
            PR=np.array([[PR]])    
            
            try:
                MD=MDCalc(X,Xold,PR)
            except:
                MD=0
            MDPr=MD_ProbCalc(MD,DF)
            MDout.append(MD)
            MD_Prob.append(MDPr)
        
        #Find changes in underlying trend
        EVMT=[]
        for i in range(self.SeasonalityPeriods+1):
            EVMT.append(0)
        ChangePoints=[]
        ChangePoint=self.SeasonalityPeriods+2
#         ChangePoint=0
        PrLmOld=0
        for t in range(self.SeasonalityPeriods+3,len(MDout)+1):
            EV=0
            EV_Old=1  #Prl1=1 when t=1
            #Iterate over a window expanding back to the last change point 

            for m in range(1,t-ChangePoint+1):  
                loc=sqrt(2*log(m+1))-(2*pi+log(log(m+1)))/(2*sqrt(log(m+1)))
                scale=1/sqrt(2*log(m+1))
                if m<=2:
                    EV_conditionalm=exp(-exp(-(MDout[t-1])))
                else:
                    EV_conditionalm=exp(-exp(-(MDout[t-1]-loc)/scale))
        
                if m==1:
                    PrLm=EV_Old
                else:
                    PrLm=(1-EV_Old)*PrLmOld
        
                EV+=EV_conditionalm*PrLm 
                PrLmOld=PrLm
                EV_Old=EV
            #Changepoint if likelihood > 95% and not in first few datapoints and no changepoints in last 3 periods
            if   EV>0.97 and t>self.SeasonalityPeriods*2+2 and (t-(ChangePoint+1))>2:  
                EV=1
                ChangePoint=t
                PrLmOld=0
                EV_Old=1
                ChangePoints.append(ChangePoint)
            EVMT.append(EV)
            
        ExtendedDates=list(self.ts.dates)
        #Do forecast if no change points in last 3 periods
        if  not sum([i > len(self.ts.data)-4 for i in ChangePoints]):
            #Start with historical data
            Prediction=self.TimeseriesModel.KalmanModelFit.predict(0,len(self.ts.data))
            
            #Take average of last 3 growth factors or mostrecent, whatever is smaller
            Trend=[self.TimeseriesModel.KalmanModelFit.filtered_state[-1,x] \
                   for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
            if abs(Trend[-1])<abs(sum(Trend[-3:])/3):
                AverageGrowth=Trend[-1]
            else:
                AverageGrowth=sum(Trend[-3:])/3                 
            
            #Get latest historical state
            LastState=self.TimeseriesModel.KalmanModelFit.filtered_state[:,-1]
            #Replace actual growth with average growth
            LastState[-1]=AverageGrowth
            
            
            Transition=self.TimeseriesModel.StateArrays.transition
            Design=self.TimeseriesModel.StateArrays.design
            #Add forecast to historical data
            for i in range(self.ForecastPeriods):
                    LastState=Transition@LastState
                    Prediction=np.append(Prediction,Design@LastState)

            #add extra dates
            
            Period= self.ts.Period
            for i in range(0,self.ForecastPeriods):
                if Period =="months":
                    NextDate=self.ts.dates[-1]+relativedelta(months=i+1)
                elif Period == 'weeks':
                    NextDate=self.ts.dates[-1]+relativedelta(weeks=i+1)
                else:
                    NextDate=self.ts.dates[-1]+relativedelta(days=i+1)
                ExtendedDates.append(NextDate)
                
            PredictionVariance=self.TimeseriesModel.KalmanModelFit.get_prediction(0,len(self.ts.data)+\
                                                                                  self.ForecastPeriods).var_pred_mean
        else:
            Prediction=self.TimeseriesModel.KalmanModelFit.predict(0,len(self.ts.data))
            PredictionVariance=self.TimeseriesModel.KalmanModelFit.get_prediction(0,len(self.ts.data)).var_pred_mean
            ExtendedDates=self.ts.dates
        
        Prediction=Prediction*np.linalg.norm(self.ts.data)
        PredictionVariance=PredictionVariance*(np.linalg.norm(self.ts.data)**1)
        
        Trend=[self.TimeseriesModel.KalmanModelFit.filtered_state[-1,x] \
               for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
        Trend=[x*np.linalg.norm(self.ts.data) for x in Trend]
        TrendSmoothed=[self.TimeseriesModel.KalmanModelFit.smoothed_state[-1,x] \
               for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
        TrendSmoothed=[x*np.linalg.norm(self.ts.data) for x in TrendSmoothed]
        
        TrendVariance=[self.TimeseriesModel.KalmanModelFit.filtered_state_cov[-1,-1,x] \
                       for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state_cov.shape[2])]
        TrendVariance=[x*np.linalg.norm(self.ts.data)**1 for x in TrendVariance]
        
        for i in range(self.SeasonalityPeriods+2):
            PredictionVariance[i]=0
            TrendVariance[i]=0

        
        ChangePoints=[1 if x in ChangePoints else 0 for x in range(1,len(gs_orig)+1)]
        
        Results= ModuleResults(Trend=Trend,TrendSmoothed=TrendSmoothed,TrendVariance=TrendVariance,ChangePoints=ChangePoints,Prediction=Prediction,\
                              PredictionVariance=PredictionVariance,ExtendedDates=ExtendedDates)
        
        
        return Results

==> ChangePointDetectorEVT-0.0.17.tar.gz: ChangePointDetectorEVT-0.0.17/src/ChangePointDetector/ChangePointDetector.py <==
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 17:56:11 2021
v0.8 28Oct21

@author: Michael Hauptman

CHANGE POINT DETECTOR

This module takes a time series and returns the uunderlying linear trend and changpoints. It uses EVT as described in https://www.robots.ox.ac.uk/~sjrob/Pubs/LeeRoberts_EVT.pdf

INSTRUCTIONS:

1. from ChangePointDetector import ChangePointDetector 
2. Prepare your time series as data plus Panda dates
3. Create  the necessary Kalman representation by creating a "session" object by calling the ChangePoint class, e.g.:
	Session=ChangePointDetector.ChangePointDetectorSession(data,dates). 
    - 'SeasonalityPeriods' is an optional input, e.g 12 = calendar month seasonality
    - 'ForecastPeriods' is another optional input, indicating how many periods to forecast.  Default = 3
4. Determine the changepoints by running the ChangePointDetectorFunction on your "session", e.g.
	Results=Session.ChangePointDetectorFunction()
5. This will return a "Results" object that contains the following:
	- ChangePoints.  This is a list of 0s and 1s the length of the data, where 1s represent changepoints
	- Prediction.  This is the Kalman smoothed actuals, plus a 3 period forecast.  Note no forecast will be made if there is a changepoint in the last 3 		dates
	- PredictionVariance
	- ExtendedDates.  These are the original dates plus 3 exta for the forecast (if a forecast has been made)
	- Trend.  This is the linear change factor
	- TrendVariance.  Variance of the trend


"""

import numpy as np
from numpy.linalg import inv
from statsmodels.tsa.statespace.mlemodel import MLEModel
# import math
from math import sqrt, log, exp,pi
from scipy.stats.distributions import chi2
from dateutil.relativedelta import relativedelta

class ModuleResults:
    def __init__(self,Trend,TrendSmoothed,TrendVariance,ChangePoints,Prediction,PredictionVariance,ExtendedDates):
            self.Trend=Trend
            self.TrendSmoothed=TrendSmoothed
            self.TrendVariance=TrendVariance
            self.ChangePoints=ChangePoints
            self.Prediction=Prediction
            self.ExtendedDates=ExtendedDates
            self.PredictionVariance=PredictionVariance

class TimeSeriesData:
        def __init__(self,data,dates):
            self.data=data
            self.dates=dates
            #Determine periodicity
            Delta=(dates[1]-dates[0])
            if Delta in range(27,32):
                Period = "months"
            elif Delta ==1:
                Period = 'days'
            elif Delta ==7:
                Period = 'weeks'
            else:
                Period ="months"
            self.Period=Period

            
class SeasonalStateArrays:
    def __init__(self,SeasonalityPeriods):
        if SeasonalityPeriods ==0:
            A=np.diag(np.ones(1))
            H=np.ones(1)
        elif SeasonalityPeriods ==1:
            A=np.array([[1,1],[0,1]])
            H=np.array([1,0])
        else:
            A=np.diag(np.ones(SeasonalityPeriods+1))
            B=np.zeros((SeasonalityPeriods+1,1))
            C=-1*np.ones(SeasonalityPeriods+2)
            C[0]=0
            C[-1]=0
            A=np.hstack([A,B])
            A[0,-1]=1
            D=np.zeros(SeasonalityPeriods+2)
            D[-1]=1
            A[-1,:]=D
            A=np.vstack([A[0,:],C,A[1:,:]]).astype('float')
            H=np.zeros((SeasonalityPeriods+2))
            H[0]=1
            H[1]=1
        
        self.transition=A
        self.design=np.array([H])
        
class KalmanModel(MLEModel):
    def __init__(self, endog, transition,design,**dates):
        super().__init__(endog, k_states = transition.shape[0], **dates)

        self['transition']=transition
        self['design']=design
        self['selection']=np.eye(transition.shape[0])

        self.initialize_approximate_diffuse()

    def update(self, params, **kwargs):
        params = super().update(params, **kwargs)
        self['state_cov']=np.eye(self.k_states)
        for i in range(self.k_states):
            self['state_cov'][i,i]=params[0]
        for j in range(self['design'].shape[0]):
            self['obs_cov',j, j] = params[1] ** 2
            
def kgain(P, H, R):
    IS =  H @ P @ H.T + R # the Covariance of predictive mean of Y
    K = P @ H.T @ inv(IS) #State covariance in observation space divided by  Covariance of predictive mean of Y
    return K

def MDCalc(X,Xold,P):
        return(
        sqrt((X-Xold).T@inv(P)@(X-Xold))  #Mahalanobis distance
        )
    
def MD_ProbCalc(MD,DF):
    return(
    chi2.cdf(MD, df=DF)
    )

class MyKalmanFilter:
    def __init__(self, TS_Data, SeasonalityPeriods=1):
        OriginalData=TS_Data.data
        NormalisedData=OriginalData/np.linalg.norm(OriginalData) 
#         NormalisedData=RemoveOutliers(OriginalData)/np.linalg.norm(RemoveOutliers(OriginalData))
        self.OriginalData=OriginalData
        self.data=NormalisedData
        self.dates=TS_Data.dates
        self.SeasonalityPeriods=SeasonalityPeriods
        self.StateArrays=SeasonalStateArrays(SeasonalityPeriods)
            
        
    def InitialiseFilter(self):
        NormalisedData=self.data
        MeasureVar=.01   #Doesn't seem to do anything??
        kf_model=KalmanModel(NormalisedData,transition=self.StateArrays.transition,design=self.StateArrays.design)
        
        start_params = [np.var(NormalisedData)**3, MeasureVar] #State convariance and observation covariance. These settings will underfit the curve
        kf_model_fit = kf_model.fit(start_params,maxiter = 20,method = 'bfgs', hessian= 'true')
        Growth=kf_model_fit.predict(1,len(self.data))
        Base=np.mean(abs(NormalisedData-Growth))

        Success=False
        i=3
        OldRatio=0
        while Success ==False and i>0:  #Increae state covariance until fitting improves over underfit base case
            start_params = [np.var(NormalisedData)**(i), MeasureVar]
            kf_model_fit = kf_model.fit(start_params,maxiter = 20,method = 'bfgs', hessian= 'true')
            Growth=kf_model_fit.predict(1,len(self.data))
            Ratio=np.mean(abs(NormalisedData-Growth))/Base
            if Ratio<.3 or Ratio==OldRatio:  
                Success =True
            i-=0.05
            OldRatio=Ratio
        
        #Determine kalman gain
#         K_gain=np.zeros(self.SeasonalityPeriods+1)
#         for i in range(kf_model_fit.filtered_state.shape[1]):
#             P=kf_model_fit.filtered_state_cov[:,:,i]
#             KGain=kgain(P, kf_model['design'], kf_model['obs_cov'])
#             K_gain=np.vstack([K_gain,KGain.reshape([1,self.SeasonalityPeriods+1])[0]])
            
        self.KalmanModel=kf_model
        self.KalmanModelFit=kf_model_fit
#         self.K_gain=K_gain
        

   
class ChangePointDetectorSession:
    def __init__(self, data,dates, SeasonalityPeriods=1,ForecastPeriods=3):
        self.ts=TimeSeriesData(data,dates)
        self.TimeseriesModel = MyKalmanFilter(self.ts,SeasonalityPeriods)
        self.TimeseriesModel.InitialiseFilter()
        self.SeasonalityPeriods=SeasonalityPeriods
        self.ForecastPeriods=ForecastPeriods
        
    
    def ChangePointDetectorFunction(self):
    #This is the main function    
        #Get underlying trend
#       gs_orig=self.TimeseriesModel.KalmanModelFit.filtered_state[-1,:]
        gs_orig=self.TimeseriesModel.KalmanModelFit.smoothed_state[-1,:]
        gs_orig= np.insert(gs_orig[:-1], 0, 0, axis=0)
        
        gs=TimeSeriesData(gs_orig,self.ts.dates)
        OnePeriodRegressionModel=MyKalmanFilter(gs,SeasonalityPeriods=0)
        OnePeriodRegressionModel.InitialiseFilter()
       
        
        MDout,MD_Prob = [],[]
        DF=1
        for i in range(OnePeriodRegressionModel.KalmanModelFit.filtered_state.shape[1]):
        
            X=OnePeriodRegressionModel.KalmanModelFit.filtered_state[:,i]
            P=OnePeriodRegressionModel.KalmanModelFit.filtered_state_cov[:,:,i]
            if i==0:
                Xold=X
            else:
                Xold=OnePeriodRegressionModel.KalmanModelFit.filtered_state[:,i-1]
            
            #If variance = 0, adjust to be >0 else MD cannot be calculated
            if P[0][0]==0:
                PR=X[0]/1000
            else:
                PR=P[0][0]
            PR=np.array([[PR]])    
            
            try:
                MD=MDCalc(X,Xold,PR)
            except:
                MD=0
            MDPr=MD_ProbCalc(MD,DF)
            MDout.append(MD)
            MD_Prob.append(MDPr)
        
        #Find changes in underlying trend
        EVMT=[]
        for i in range(self.SeasonalityPeriods+1):
            EVMT.append(0)
        ChangePoints=[]
        ChangePoint=self.SeasonalityPeriods+2
#         ChangePoint=0
        PrLmOld=0
        for t in range(self.SeasonalityPeriods+3,len(MDout)+1):
            EV=0
            EV_Old=1  #Prl1=1 when t=1
            #Iterate over a window expanding back to the last change point 

            for m in range(1,t-ChangePoint+1):  
                loc=sqrt(2*log(m+1))-(2*pi+log(log(m+1)))/(2*sqrt(log(m+1)))
                scale=1/sqrt(2*log(m+1))
                if m<=2:
                    EV_conditionalm=exp(-exp(-(MDout[t-1])))
                else:
                    EV_conditionalm=exp(-exp(-(MDout[t-1]-loc)/scale))
        
                if m==1:
                    PrLm=EV_Old
                else:
                    PrLm=(1-EV_Old)*PrLmOld
        
                EV+=EV_conditionalm*PrLm 
                PrLmOld=PrLm
                EV_Old=EV
            #Changepoint if likelihood > 95% and not in first few datapoints and no changepoints in last 3 periods
            if   EV>0.97 and t>self.SeasonalityPeriods*2+2 and (t-(ChangePoint+1))>2:  
                EV=1
                ChangePoint=t
                PrLmOld=0
                EV_Old=1
                ChangePoints.append(ChangePoint)
            EVMT.append(EV)
            
        ExtendedDates=list(self.ts.dates)
        #Do forecast if no change points in last 3 periods
        if  not sum([i > len(self.ts.data)-4 for i in ChangePoints]):
            #Start with historical data
            Prediction=self.TimeseriesModel.KalmanModelFit.predict(0,len(self.ts.data))
            
            #Take average of last 3 growth factors or mostrecent, whatever is smaller
            Trend=[self.TimeseriesModel.KalmanModelFit.filtered_state[-1,x] \
                   for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
            if abs(Trend[-1])<abs(sum(Trend[-3:])/3):
                AverageGrowth=Trend[-1]
            else:
                AverageGrowth=sum(Trend[-3:])/3                 
            
            #Get latest historical state
            LastState=self.TimeseriesModel.KalmanModelFit.filtered_state[:,-1]
            #Replace actual growth with average growth
            LastState[-1]=AverageGrowth
            
            
            Transition=self.TimeseriesModel.StateArrays.transition
            Design=self.TimeseriesModel.StateArrays.design
            #Add forecast to historical data
            for i in range(self.ForecastPeriods):
                    LastState=Transition@LastState
                    Prediction=np.append(Prediction,Design@LastState)

            #add extra dates
            
            Period= self.ts.Period
            for i in range(0,self.ForecastPeriods):
                if Period =="months":
                    NextDate=self.ts.dates[-1]+relativedelta(months=i+1)
                elif Period == 'weeks':
                    NextDate=self.ts.dates[-1]+relativedelta(weeks=i+1)
                else:
                    NextDate=self.ts.dates[-1]+relativedelta(days=i+1)
                ExtendedDates.append(NextDate)
                
            PredictionVariance=self.TimeseriesModel.KalmanModelFit.get_prediction(0,len(self.ts.data)+\
                                                                                  self.ForecastPeriods).var_pred_mean
        else:
            Prediction=self.TimeseriesModel.KalmanModelFit.predict(0,len(self.ts.data))
            PredictionVariance=self.TimeseriesModel.KalmanModelFit.get_prediction(0,len(self.ts.data)).var_pred_mean
            ExtendedDates=self.ts.dates
        
        Prediction=Prediction*np.linalg.norm(self.ts.data)
        PredictionVariance=PredictionVariance*(np.linalg.norm(self.ts.data)**1)
        
        Trend=[self.TimeseriesModel.KalmanModelFit.filtered_state[-1,x] \
               for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
        Trend=[x*np.linalg.norm(self.ts.data) for x in Trend]
        TrendSmoothed=[self.TimeseriesModel.KalmanModelFit.smoothed_state[-1,x] \
               for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state.shape[1])]
        TrendSmoothed=[x*np.linalg.norm(self.ts.data) for x in TrendSmoothed]
        
        TrendVariance=[self.TimeseriesModel.KalmanModelFit.filtered_state_cov[-1,-1,x] \
                       for x in range(self.TimeseriesModel.KalmanModelFit.filtered_state_cov.shape[2])]
        TrendVariance=[x*np.linalg.norm(self.ts.data)**1 for x in TrendVariance]
        
        for i in range(self.SeasonalityPeriods+2):
            PredictionVariance[i]=0
            TrendVariance[i]=0

        
        ChangePoints=[1 if x in ChangePoints else 0 for x in range(1,len(gs_orig)+1)]
        
        Results= ModuleResults(Trend=Trend,TrendSmoothed=TrendSmoothed,TrendVariance=TrendVariance,ChangePoints=ChangePoints,Prediction=Prediction,\
                              PredictionVariance=PredictionVariance,ExtendedDates=ExtendedDates)
        
        
        return Results
